/*
 * Ratiometric Frequency Counter
 *
 * displays 6 digits from 1Hz up to ~40MHz
 * cycle time 4M/F_CPU ... 16M/F_CPU
 * theoretical lower frequency limit: F_CPU/16M
 * theoretical upper frequency limit: 4*F_CPU
 *
 * (c) Axel (XL) Schwenke, axel.schwenke@gmx.net
 *
 * $Id: main.c 126 2011-03-07 13:58:36Z schwenke $
 */

#if F_CPU > (16*1024*1024-512)
#error "F_CPU is too big! Must fit in 24 bits"
#endif

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include "avr_def.h"
#include "arith.h"
#include "config.h"
#include "lcd.h"


/* function prototypes */

void hw_init(void);
void hexout(uint8_t);
unsigned char calculate_frequency(void);
void SHOW_GREETING(void);
void SHOW_RESULT(void);


/* global variables */

volatile union {    /* 24 bit counter f_x */
    uint8_t  b[3];
} nx;

volatile union {    /* 24 bit counter f_ref */
    uint16_t w[1];
    uint8_t  b[3];
} nref;

volatile union {    /* 48 bit result */
    uint32_t q[1];
    uint8_t  b[6];
} res;

uint32_t fref;      /* reference frequency */
uint8_t  jumpers;   /* jumper settings */

#define BUFSIZE 10
char buffer[BUFSIZE]; /* display buffer */


/* constants in EEPROM */

EEMEM uint32_t ee_fref = F_CPU;


/* the code */

int main(void)
{
    hw_init();

    SHOW_GREETING();

#if WDT && !DEBUG
    wdt_enable(WDTO_2S);
#endif

    while (1) {

        /* clear counters */
        clear_low();
        nx.b[2]= 0;
        TCNT0= 0;
        clear_high();

        PHASE1(); /* wait for gate to open */

        TCCR1B &= ~_BV(ICES1);  /* capture falling edge */
        TIMSK1 |=  _BV(TOIE1);  /* enable Timer1 overflow irq */	// **-us
        sei();
        start_high();
        while (gate_is_closed()) {
            wdt_reset();
            sleep_cpu();
        }

        PHASE2(); /* wait for MINCYCLES fref pulses */

        while ((nref.b[2] < (MINCYCLES/65536)+1) && (nx.b[2] < 0x7c)) {
            wdt_reset();
            sleep_cpu();
        }

        PHASE3(); /* wait for gate to close */

        TCCR1B |=  _BV(ICES1);  /* capture rising edge */
        start_low();
        while (nref.b[2] && gate_is_open()) {
            wdt_reset();
        }

        cli();

        if (nref.b[2] == 0) {
            UNDERFLOW();        /* f_ref counter overflowed */

        } else {

            /* read fx counter */
            lat_oe_low();
            lat_oe_high();      /* rising edge latches 74HC590 */
            lat_oe_low();       /* low enables 74HC590 drivers */

            register union {    /* ugly workaround */
                uint16_t w;     /* to avoid even uglier gcc code */
                uint8_t  b[2];
            } tmp;

            tmp.b[0]= TCNT0;
            tmp.b[1]= nx.b[2];
            tmp.w <<= 1;
            tmp.b[0] |= fx9_read();

            nx.b[0]= data_read();
            nx.b[1]= tmp.b[0];
            nx.b[2]= tmp.b[1];
            lat_oe_high();

            /* fref = EEPROM constant + calibration */
            fref= eeprom_read_dword(&ee_fref);
            register uint32_t cal= ADC;
            if (cal != 0) {
                fref += (cal - 512);
            }

            /* read jumper/switch settings */
            jumpers= data_read();

            SHOW_RESULT();
        }
    }
}


/* Interrupt Service Routines */

/* Timer 0 Overflow - f_x counter */
ISR(TIMER0_OVF_vect)
{
    ++nx.b[2];
}

/* Timer 1 Overflow - f_ref counter */
ISR(TIMER1_OVF_vect)
{
    ++nref.b[2];
}

/* Timer 1 Capture - gate open/close */
ISR(TIMER1_CAPT_vect)
{
    if ((TCCR1B & _BV(ICES1)) == 0) {
        /* falling edge = gate opens */
        nref.w[0]= ICR1;
        nref.b[2]= 0;
    } else {
        /* rising edge = gate closes */
        register uint16_t tmp= ICR1;
        if ((TIFR1 & _BV(TOV1)) && tmp < 256) {
            /* overflow irq pending - possible collision */
            ++nref.b[2];
        }
        if (tmp < nref.w[0]) {
            --nref.b[2];
        }
        nref.w[0]= tmp - nref.w[0];
        TIMSK1 &= ~_BV(TOIE1);  /* disable and */
        TIFR1   =  _BV(TOV1);   /* clear overflow interrupt */
    }
}


/* Sub Routines */

void hw_init(void)
{
    /* pullups enabled on all inputs */
    /* data bus pins are inputs */
    /* START=L, CLR=L, LAT/OE=H */

    PORTB= ~(_BV(CLEAR_PIN) | _BV(START_PIN));
    DDRB=   (_BV(CLEAR_PIN) | _BV(START_PIN));

    PORTC= ~(_BV(ANALOG_PIN));
    DDRC=  0;

    PORTD= ~(_BV(LCD_E_PIN));
    DDRD=   (_BV(LCD_RS_PIN) | _BV(LCD_RW_PIN) | _BV(LCD_E_PIN) | _BV(LAT_OE_PIN));

    /* disable Analog Comparator */

    ACSR = (_BV(ACD));

    /* ADC senses PC5 continuously, Vref=Aref */

    ADMUX=  (_BV(MUX2) | _BV(MUX0));
    ADCSRA= (_BV(ADEN) | _BV(ADSC) | _BV(ADATE) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0));

    /* Timer 0: ext. trigger, falling edge */
    /* Timer 1: system clock */

    TCCR0B=  (_BV(CS02) | _BV(CS01));
    TCCR1A= 0;
    TCCR1B= (_BV(CS10));

    /* Timer 0: capture irq and
     * Timer 1: overflow irq enabled */

#if defined(__AVR_ATmega8__)
    TIMSK=  (_BV(TICIE1) | _BV(TOIE0));
#else
    TIMSK0 = _BV(TOIE0);
    TIMSK1 = _BV(ICIE1);
#endif


#if SLEEP
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_enable();
#endif

#if DEBUG
    lcd_init(LCD_DISP_ON_CURSOR);
#else
    lcd_init(LCD_DISP_ON);
#endif
}


void hexout(uint8_t x)
{
    register uint8_t d= (x>>4) & 0x0F;
    lcd_putc(d + (d>9 ? 'A'-10 : '0'));
    d= x & 0x0F;
    lcd_putc(d + (d>9 ? 'A'-10 : '0'));
}


unsigned char calculate_frequency(void)
{
    uint8_t prefix='m'; /* unit prefix 'm'Hz */
    uint8_t decimal=10; /* decimal point after digit #10 */

    /* res *= fref * nx */
    res.q[0] = fref;
    mymul((void*) &res, (void*) &nx);

    /* multiply by 10 until big enough
     * this shifts the decimal position */
    while ((res.b[5] & 0xF8) == 0 && decimal > 1) {
        /* res *= 10 */
        mymulby10((void*) &res);
        --decimal;
    }

    /* res /= nref */
    mydiv((void*) &res, (void*) &nref);

    /* res = PackedBCD(res) */
    mybin2bcd((void*) &res);

    /* skip leading zeros */
    while (decimal && (res.b[3] & 0xF0) == 0) {
        res.q[0] <<= 4;
        --decimal;
    }

    /* adjust unit prefix */
    if (decimal>2) {
        prefix   = ' ';
        decimal -= 3;
    }
    if (decimal>2) {
        prefix   = 'k';
        decimal -= 3;
    }
    if (decimal>2) {
        prefix   = 'M';
        decimal -= 3;
    }

    /* write result to display buffer */
    unsigned char len= 0;
    char *bufptr= buffer;

    /* write 6 digits + decimal separator */
    for (register uint8_t i=0; i<6; i++) {
        (*bufptr++)= ('0' + ((res.b[3] & 0xF0) >> 4));
        len++;
        res.q[0] <<= 4;
        if (i==decimal) {
            (*bufptr++)= '.';
            len++;
        }
    }

    /* append unit */
    (*bufptr++)= prefix;
    (*bufptr++)= 'H';
    (*bufptr++)= 'z';
    len+= 3;

    return len;
}

//eingefuegt Periode----------------------------------------------------------------------
unsigned char calculate_period(void)
{
    uint8_t prefix='n'; /* unit prefix 'n' for ns */
    uint8_t decimal=16; /* decimal point after digit #16 */

    union {
        uint8_t  b[4];
        uint32_t q[1];
    } nx2;

    /* normalize nx */
    nx2.b[0] = nx.b[0];
    nx2.b[1] = nx.b[1];
    nx2.b[2] = nx.b[2];
    nx2.b[3] = 0;

    while ((nx2.b[2] & 0xF8) == 0) {
        nx2.q[0] *= 10;
        ++decimal;
    }

    /* res = nref */
    res.b[0] = nref.b[0];
    res.b[1] = nref.b[1];
    res.b[2] = nref.b[2];
    res.b[3] = 0;
    res.b[4] = 0;
    res.b[5] = 0;

    /* multiply by 10 until big enough
     * this shifts the decimal position */
    while ((res.b[5] & 0xF0) == 0 && decimal > 1) {
        /* res *= 10 */
        mymulby10((void*) &res);
        --decimal;
    }

    /* res /= nx */
    mydiv((void*) &res, (void*) &nx2);

    /* multiply by 10 until big enough
     * this shifts the decimal position */
    while ((res.b[5] & 0xF0) == 0 && decimal > 1) {
        /* res *= 10 */
        mymulby10((void*) &res);
        --decimal;
    }

    /* res /= fref */
    mydiv((void*) &res, (void*) &fref);

    /* res = PackedBCD(res) */
    mybin2bcd((void*) &res);

    /* skip leading zeros */
    while (decimal && (res.b[3] & 0xF0) == 0) {
        res.q[0] <<= 4;
        --decimal;
    }

    /* adjust unit prefix */
    if (decimal>2) {
        prefix   = 'u';
        decimal -= 3;
    }
    if (decimal>2) {
        prefix   = 'm';
        decimal -= 3;
    }
    if (decimal>2) {
        prefix   = ' ';
        decimal -= 3;
    }

    /* write result to display buffer */
    unsigned char len= 0;
    char *bufptr= buffer;

    /* write 6 digits + decimal separator */
    for (register uint8_t i=0; i<6; i++) {                   // i<6;   geandert--------------------
        (*bufptr++)= ('0' + ((res.b[3] & 0xF0) >> 4));
        len++;
        res.q[0] <<= 4;
        if (i==decimal) {
            (*bufptr++)= '.';
            len++;
        }
    }

    /* append unit */
    (*bufptr++)= prefix;
    (*bufptr++)= 's';       // geandert "S" --------------------------------------------------------
    len+= 2;

    return len;
}
// eingefuegt----------------------------------------------------------------------------


/*
 * customized functions
 */

/* for display 1x10 */

void show_result_1x10(void)
{
    unsigned char len= calculate_frequency();
    char* bufptr= buffer;

    lcd_clrscr();
    while (len--) {
        lcd_putc(*bufptr++);
    }
    IS_VALID();
}


void show_greeting_1x10(void)
{
    lcd_puts(VERSION);
    _delay_ms(1000);

    /* display current reference freq */
    res.q[0]= eeprom_read_dword(&ee_fref);
    mybin2bcd((void*) &res);
    lcd_clrscr();
    lcd_putc('R');
    lcd_putc('=');
    hexout(res.b[3]);
    hexout(res.b[2]);
    hexout(res.b[1]);
    hexout(res.b[0]);
    _delay_ms(2000);

    /* display current adjustment */
    res.q[0]= ADC;
    lcd_clrscr();
    lcd_puts("ADJ=");
    if (res.q[0] == 0) {
        lcd_puts("(off)");
    } else {
        if (res.q[0] >= 512) {
            lcd_putc('+');
            res.q[0] -= 512;
        } else {
            lcd_putc('-');
            res.q[0] = 512 - res.q[0];
        }
        mybin2bcd((void*) &res);
        hexout(res.b[1]);
        hexout(res.b[0]);
    }
    _delay_ms(2000);
}


/* for display 1x16 */

void show_result_1x16(void)
{
    unsigned char len= calculate_frequency();
    char* bufptr= buffer;

    lcd_clrscr();
    lcd_putc('F');
    lcd_putc('=');
    while (len--) {
        lcd_putc(*bufptr++);
    }
    IS_VALID();
}


void show_greeting_1x16(void)
{
    lcd_puts(VERSION);
    _delay_ms(1000);

    /* display current reference freq */
    res.q[0]= eeprom_read_dword(&ee_fref);
    mybin2bcd((void*) &res);
    lcd_clrscr();
    lcd_putc('R');
    lcd_putc('=');
    hexout(res.b[3]);
    hexout(res.b[2]);
    hexout(res.b[1]);
    hexout(res.b[0]);

    /* display current adjustment */
    res.q[0]= ADC;
    if (res.q[0] != 0) {
        if (res.q[0] >= 512) {
            lcd_putc('+');
            res.q[0] -= 512;
        } else {
            lcd_putc('-');
            res.q[0] = 512 - res.q[0];
        }
        mybin2bcd((void*) &res);
        hexout(res.b[1]);
        hexout(res.b[0]);
    }
    _delay_ms(2000);
}


/* for display 2x16 */

void show_result_de0508_2x16(void)                                            
{                                                                              
    unsigned char len= calculate_frequency();                                  
    char* bufptr= buffer;                                                      
                                                                               
    lcd_clrscr();                                                              
    lcd_putc('F');                                                             
    lcd_putc('~');
/* Trennzeichen nach Geschmack eingefuegt----------------------------2. Leerzeichen nach "F" 1.DisplayZeile----------*/                                        
    lcd_putc(' ');                                                             
                                                                               
    /* nur die Zahl ausgeben */                                                
    while ((*bufptr == '.' || (*bufptr >= '0' && *bufptr <= '9')) && len--) {  
        lcd_putc(*bufptr++);                                                   
    }                                                                          
                                                                               
    /* Trennzeichen nach Geschmack eingefuegt------------------------------------------------ */                                          
    lcd_putc(' ');                                                             
                                                                               
    /* und jetzt noch die Einheit  eingefuegt-------------------------------------------------*/                                           
    while (len--) {                                                            
        lcd_putc(*bufptr++);                                                   
    }                                                                          
                                                                               
    len= calculate_period();
    bufptr= buffer;
    lcd_puts("\nT~");
	/* Trennzeichen nach Geschmack eingefuegt------------------------2. Leerzeichen nach "T" 2.DisplayZeile----------*/                                          
    lcd_putc(' ');
	/* nur die Zahl ausgeben       eingefuegt-------------------------------------------------*/                                                
    while ((*bufptr == '.' || (*bufptr >= '0' && *bufptr <= '9')) && len--) {  
        lcd_putc(*bufptr++);                                                   
    }
	/* Trennzeichen nach Geschmack eingefuegt-------------------------------------------------*/                                          
    lcd_putc(' ');
	
    while (len--) {
        lcd_putc(*bufptr++);
    }

    IS_VALID();
}

void show_greeting_de0508_2x16(void)
                                          
{                                                                              
    lcd_puts(VERSION"\nHallo Michael!");                                       
    _delay_ms(2000);                                                           
 /* display current reference freq */
    res.q[0]= eeprom_read_dword(&ee_fref);
    mybin2bcd((void*) &res);
    lcd_puts("\nR=");
    hexout(res.b[3]);
    hexout(res.b[2]);
    hexout(res.b[1]);
    hexout(res.b[0]);

    /* display current adjustment */
    res.q[0]= ADC;
    if (res.q[0] != 0) {
        if (res.q[0] >= 512) {
            lcd_putc('+');
            res.q[0] -= 512;
        } else {
            lcd_putc('-');
            res.q[0] = 512 - res.q[0];
        }
        mybin2bcd((void*) &res);
        hexout(res.b[1]);
        hexout(res.b[0]);
    }
    _delay_ms(2000);


} 

void show_result_2x16(void)
{
    unsigned char len= calculate_frequency();
    char* bufptr= buffer;

    lcd_clrscr();
    lcd_putc('F');
    lcd_putc('=');

    while (len--) {
        lcd_putc(*bufptr++);
    }
    IS_VALID();
}


void show_greeting_2x16(void)
{
    lcd_puts(VERSION);

    /* display current reference freq */
    res.q[0]= eeprom_read_dword(&ee_fref);
    mybin2bcd((void*) &res);
    lcd_puts("\nR=");
    hexout(res.b[3]);
    hexout(res.b[2]);
    hexout(res.b[1]);
    hexout(res.b[0]);

    /* display current adjustment */
    res.q[0]= ADC;
    if (res.q[0] != 0) {
        if (res.q[0] >= 512) {
            lcd_putc('+');
            res.q[0] -= 512;
        } else {
            lcd_putc('-');
            res.q[0] = 512 - res.q[0];
        }
        mybin2bcd((void*) &res);
        hexout(res.b[1]);
        hexout(res.b[0]);
    }
    _delay_ms(2000);
}


/* debug version 2x16 */

void show_result_debug(void)
{
    unsigned char len= calculate_frequency();
    char* bufptr= buffer;

    lcd_clrscr();
    hexout(nx.b[2]);
    hexout(nx.b[1]);
    hexout(nx.b[0]);
    lcd_putc('/');
    hexout(nref.b[2]);
    hexout(nref.b[1]);
    hexout(nref.b[0]);
    lcd_putc(' ');
    hexout(jumpers);

    lcd_puts("\nF=");
    while (len--) {
        lcd_putc(*bufptr++);
    }
    IS_VALID();
}


void show_greeting_debug(void)
{
    lcd_puts(VERSION);
    lcd_puts("-debug");

    /* display current reference freq */
    res.q[0]= eeprom_read_dword(&ee_fref);
    mybin2bcd((void*) &res);
    lcd_puts("\nR=");
    hexout(res.b[3]);
    hexout(res.b[2]);
    hexout(res.b[1]);
    hexout(res.b[0]);

    /* display current adjustment */
    res.q[0]= ADC;
    if (res.q[0] != 0) {
        if (res.q[0] >= 512) {
            lcd_putc('+');
            res.q[0] -= 512;
        } else {
            lcd_putc('-');
            res.q[0] = 512 - res.q[0];
        }
        mybin2bcd((void*) &res);
        hexout(res.b[1]);
        hexout(res.b[0]);
    }
    _delay_ms(2000);
}


